#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "tetris.h"
#include "game.h"
#include "options.h"
#include "input.h"
#include "socket.h"
#include "hiscore.h"
#include "menu.h"

int testgameopt(const char *key, int val, int pl)
{
	const char *keys[4] = {"level", "height", "lines", "rotate"};
	const char max[4] = {9, 5, 127, 3};
	int i;
	if (val < 0)
		return 0;
	if (!pl) {
		if (!strcmp(key, "mode"))
			return 1;
		else if (!strcmp(key, "softdrop"))
			return 6;
	}
	for (i = 0; i < 4; i++) {
		if (!strcmp(key, keys[i]))
			return val <= max[i] ? i+2 : 0;
	}
	return 0;
}

static void initplayer(struct player *p)
{
	p->startlevel = 0;
	p->height = 0;
	p->lineslimit = 25;
	p->rotationsys = ROT_CLOCKWISE | ROT_LEFTHAND;
	p->score = 0;
}

#if defined(TWOPLAYER) && defined(JOYSTICK)
static void initplayerinput(int *devs)
{
	const char *s;
	union val v;
	inputdevs_player[0] = 0;
	if (devs[0] != devs[1]) {
		if (devs[0]) {
			inputdevs_player[devs[0]] = 1;
			if (!devs[1])
				inputdevs_player[0] = 2;
		}
		if (devs[1])
			inputdevs_player[devs[1]] = 2;
	} else if (devs[0]) {
		unsetoption("player1", "input");
		inputdevs_player[devs[0]] = 2;
	} else if (s = getinputdev_str(1)) {
		strncpy(v.str, s, 4);
		setoption("player2", "input", v, 1);
		inputdevs_player[getinputdev(s)] = 2;
	}
}
#else
#define initplayerinput(devs)
#endif

void initgame()
{
	struct player *p = game->player;
	int pl = 0;
	struct option *o = getoptions("");
	int v;
	int i;
#if defined(TWOPLAYER) && defined(JOYSTICK)
	int devs[2] = {0};
#endif
	srand(time(NULL));
	game->mode = getopt_int("", "mode");
	if (socket_fd > -1)
		game->mode |= MODE_2PLAYER | MODE_NETWORK;
	game->state = 0;
	game->next = NULL;
	initplayer(p);
	if (!TWOPLAYER_MODE)
		softdrop_speed = 1;
	else {
		initplayer(p+1);
		pl = 1;
		if (socket_fd == -1)
			o = getoptions("player1");
	}
	while (1) {
		if (!o) {
			if (!TWOPLAYER_MODE || pl==2 || socket_fd > -1)
				break;
			p++;
			pl = 2;
			o = getoptions("player2");
			continue;
		}
		v = o->val.integ;
		i = testgameopt(opt_key(o), v, pl);
		if (i>=2 && i<=5)
			(&p->startlevel)[i-2] = v;
		else if (i==6)
			softdrop_speed = v;
#if defined(TWOPLAYER) && defined(JOYSTICK)
		else if (TWOPLAYER_MODE && !strcmp(opt_key(o), "input"))
			devs[pl-1] = getinputdev(o->val.str);
#endif
		o = o->next;
	}
	if (socket_fd > -1)
		sock_initgame();
	else if (TWOPLAYER_MODE)
		initplayerinput(devs);
}

int startgame()
{
	setupplayer(&player1);
	if (TWOPLAYER_MODE)
		setupplayer(&player2);
	else if (!hiscores[0].score)
		readhiscores(NULL);
	if (TWOPLAYER_MODE) {
		startgame_2p();
		return 0;
	}
	return startgame_1p() && gameovermenu();
}
