#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "term.h"
#include "xterm.h"
#include "options.h"
#include "draw.h"
#include "tetris.h"
#include "game.h"
#include "menu.h"
#include "configfile.h"
#include "hiscore.h"
#include "socket.h"

char term_width = 80;
char term_height = 20;

void gettermsize()
{
#ifdef TIOCGSIZE
	struct ttysize tty;
#elif defined(TIOCGWINSZ)
	struct winsize win;
#endif
	char *s;
	int w = 80;
	int h = 20;
#ifdef TIOCGSIZE
	if (!ioctl(STDIN_FILENO, TIOCGSIZE, &tty)) {
		w = tty.ts_cols;
		h = tty.ts_lines;
	}
#elif defined(TIOCGWINSZ)
	if (!ioctl(STDIN_FILENO, TIOCGWINSZ, &win)) {
		w = win.ws_col;
		h = win.ws_row;
	}
#else
	if (0) ;
#endif
	else {
		if (s = getenv("COLUMNS"))
			sscanf(s, "%d", &w);
		if (s = getenv("LINES"))
			sscanf(s, "%d", &h);
	}
	settermwidth(w);
	settermheight(h);
}

void settermwidth(int w)
{
	if (w < 32)
		w = 32;
	else if (w > 127)
		w = 127;
	term_width = w;
}

void settermheight(int h)
{
	if (h < 20)
		h = 20;
	else if (h > 25)
		h = 25;
	term_height = h;
	if (term_height < 24)
		term_flags &= ~HEIGHT_24L;
	else
		term_flags |= HEIGHT_24L;
}

static void gettermenv()
{
	char *s = getenv("TERM");
	if (s) {
#ifdef XTERM
		if (!strncmp(s, "xterm", 5))
			term_flags |= XTERM;
#endif
		else if (!strcmp(s, "linux"))
			term_flags |= LINUX_TERM;
		else if (!strcmp(s, "cygwin"))
			term_flags |= CYGWIN;
	}
	s = getenv("COLORTERM");
	if (s && !strcmp(s, "gnome-terminal"))
		term_flags |= GNOME_TERM;
}

static void gettermoptions()
{
	struct option *o;
	char *k;
	union val v;
	int bg = 0;
	int i;
	gettermenv();
	o = getoptions("term");
	for (; o; o = o->next) {
		k = opt_key(o);
		v = o->val;
		if (!strcmp(k, "bg")) {
			if (v.integ == 1)
				term_flags |= WHITE_BG;
			bg = 1;
		} else if (!strcmp(k, "drawing")) {
			if (v.integ)
				term_flags |= ASCII;
			else
				term_flags &= ~ASCII;
		} else if (!strcmp(k, "color")) {
			if (!v.integ)
				term_flags |= MONOCHROME;
		} else if (!strcmp(k, "block")) {
			if (opt_isint(o)) {
				if (v.integ == -1) {
					term_flags |= TT_BLOCKS;
					continue;
				} else if (v.integ == -2) {
					term_flags |= TT_BLOCKS_BG;
					continue;
				}
				while (abs(v.integ) > 99)
					v.integ /= 10;
				sprintf(v.str, "%d", v.integ);
			}
			for (i = 0; i < 2; i++) {
				block_chars[i] = isprintable(v.str[i])
							? v.str[i] : ' ';
				if (v.str[2] == 'a')
					block_chars[i] |= 0x100;
			}
		} else if (!strcmp(k, "bgdot")) {
			if (v.str[0] == '~')
				bgdot = BULLET;
			else if (isprintable(v.str[0]))
				bgdot = v.str[0];
			else
				bgdot = ' ';
		}
	}
	if ((term_flags & (WHITE_BG | TT_MONO))==(WHITE_BG | TT_BLOCKS))
		term_flags ^= TT_BLOCKS | TT_BLOCKS_BG;
	if (term_flags & CYGWIN)
		term_flags |= ASCII;
	if ((term_flags & (MONOCHROME | ASCII))==MONOCHROME &&
	    !getopt_int("term", "block")) {
		block_chars[0] = TEXTURE2;
		block_chars[1] = TEXTURE1;
	}
}

int getmargin_x()
{
	int x;
	if (term_width <= 32)
		x = 0;
#ifdef TWOPLAYER
	else if (!in_menu && TWOPLAYER_MODE) {
		if (term_width < 58)
			x = 0;
		else
			x = (term_width-56)/2;
	}
#endif
	else if (term_width < 48)
		x = 1;
	else
		x = (term_width-(in_menu ? 34 : 44))/2;
	return x;
}

void getwin_xy(int win, int *x, int *y)
{
	*y = 0;
	switch (win) {
	case 0:
		*x = 0;
		return;
	case 1:
	case 2:
		*x = board_x(win, 0);
		break;
	case WIN_NEXT:
	case WIN_NEXT+1:
		next_xy(1+win-WIN_NEXT, x, y);
		return;
	case WIN_PANEL:
		*x = TWOPLAYER_MODE ? 22 : 0;
		break;
	case WIN_TETROM_STATS:
		*y = 11;
	case WIN_TOP_SCORES:
		*x = board_x(1, 12);
		if (!*y && !_HEIGHT_24L)
			*y = 1;
	}
	if (_HEIGHT_24L)
		*y += 4;
}

void putch_ascii(int ch)
{
	switch (ch) {
	case HLINE:
		ch = '-';
		break;
	case VLINE:
		ch = '|';
		break;
	case TEXTURE1:
		ch = '#';
		break;
	case TEXTURE2:
	case ' '|0x100:
		ch = ' ';
		break;
	case BULLET:
		ch = '.';
		break;
	case UPARROW:
		ch = '^';
		break;
	default:
		ch = '+';
	}
	putch(ch);
}

int putnchars(int ch, int n)
{
	int i = 0;
	for (; i < n; i++)
		putch(ch);
	return n;
}

void printstr_acs(const char *str, int n)
{
	int ch = 0;
	while (*str) {
		if (islower(*str) || (ch & 0x100) && *str == ' ')
			ch = *str | 0x100;
		else
			ch = *str;
		putch(ch);
		str++;
		if (*str == 'N') {
			putnchars(ch, n-1);
			str++;
		}
	}
}

int isprintable(char c)
{
	return (c < 0 || c >= ' ' && c != 0x7F);
}

#ifndef NO_MENU
void print_vitetris_ver(int x, int y)
{
	setcurs(x, y);
	setattr_bold();
	printstr(VITETRIS_VER);
	setattr_normal();
}

static int getblockstyle()
{
	if (_TT_BLOCKS)
		return 1;
	if (_TT_BLOCKS_BG)
		return 2;
	if (!getopt_int("term", "block"))
		return 0;
	return 3;
}

static void reset_block_chars()
{
	if ((term_flags & (MONOCHROME | ASCII))==MONOCHROME) {
		block_chars[0] = TEXTURE2;
		block_chars[1] = TEXTURE1;
	} else {
		block_chars[0] = TEXTURE1;
		block_chars[1] = TEXTURE2;
	}
}

int term_optionhandler(int k, const struct termopt *o)
{
	int flag = o->flag;
	union val v;
	if (k == 1)
		term_flags &= ~flag;
	else if (k == 2)
		term_flags |= flag;
	else if (k)
		return 0;
	printmenuitem_options(o->s, (term_flags & flag)!=0);
	if (!k)
		return 1;
	v.integ = flag==MONOCHROME ? 2-k : k-1;
	setoption("term", o->key, v, 0);
	if ((term_flags & (WHITE_BG | TT_MONO))==(WHITE_BG | TT_BLOCKS) ||
	    _TT_BLOCKS_BG && (!_WHITE_BG || _MONOCHROME) &&
	    getopt_int("term", "block") == -1)
		term_flags ^= TT_BLOCKS | TT_BLOCKS_BG;
	if (flag & (MONOCHROME | ASCII) && !getblockstyle())
		reset_block_chars();
	return 3;
}

static void setblockstyle(int i, int c)
{
	union val v;
	int tp = 0;
	if (_TT_BLOCKS && bgdot==' ') {
		bgdot = default_bgdot();
		unsetoption("term", "bgdot");
	}
	term_flags &= ~(TT_BLOCKS | TT_BLOCKS_BG);
	v.integ = 0;
	switch (i) {
	case 0:
		reset_block_chars();
		unsetoption("term", "block");
		return;
	case 1:
		term_flags |= TT_BLOCKS;
		bgdot = ' ';
		setoption("term", "bgdot", v, 1);
		v.integ = -1;
		break;
	case 2:
		term_flags |= TT_BLOCKS_BG;
		v.integ = -2;
		break;
	case 3:
		block_chars[0] = c;
		block_chars[1] = ' ';
		v.str[0] = c;
		tp = 1;
	}
	setoption("term", "block", v, tp);
}

static void save_custom_blockstyle()
{
	union val v;
	v.str[0] = block_chars[0];
	if (v.str[0]==' ')
		v.str[0]--;
	v.str[1] = block_chars[1];
	if (block_chars[0] & 0x100) {
		v.str[2] = 'a';
		v.str[3] = '\0';
	} else
		v.str[2] = '\0';
	setoption("term", "block", v, 1);
}

static int edit_custom_blockstyle(int i, int c)
{
	static int pos;
	if (i != 3) {
		setblockstyle(3, c);
		pos = 1;
		return 1;
	}
	if (!pos || pos == 2 && c != 'a') {
		block_chars[0] = c;
		block_chars[1] = ' ';
		pos = 1;
		return 0;
	}
	if (pos == 1) {
		block_chars[1] = c;
		pos = _ASCII ? 0 : 2;
	} else {
		block_chars[0] |= 0x100;
		block_chars[1] |= 0x100;
		pos = 0;
	}
	save_custom_blockstyle();
	return 1;
}

int term_blockstyle(int k)
{
	const char *items[4] = {"default", "tt", "tt-bg", "custom"};
	int n = 4;
	int i = getblockstyle();
	int j;
	if (isprintable(k)) {
		if (edit_custom_blockstyle(i, k))
			return 3;
		k = 0;
	}
	if ((term_flags & (WHITE_BG | MONOCHROME)) == WHITE_BG) {
		items[1] = items[2];
		items[2] = items[3];
		n--;
		if (i)
			i--;
	}
	j = i;
	k = selectitem(items, n, &i, k);
	if (i != j) {
		setblockstyle(i+(i && n==3), 'X');
		return 3;
	}
	if (k != 1)
		return k;
	if (term_flags & (TT_BLOCKS | TT_BLOCKS_BG))
		printstr("  ");
	else {
		putch(block_chars[0]);
		putch(block_chars[1]);
	}
	return 1;
}
#endif /* !NO_MENU */

static void hiscoreline()
{
	setcolorpair(RED_FG);
	putnchars('-', 9);
	setattr_normal();
}

void print_top_scores()
{
	char s[8];
	const struct hiscore *hs = hiscores;
	int pos = 0;
	int i = 1;
	if (term_width < 47 || !hs[0].score)
		return;
	setwcurs(WIN_TOP_SCORES, 0, 0);
	setcolorpair(MAGENTA_FG);
	printstr("Top Scores");
	setattr_normal();
	while (hs->score && i <= 5) {
		newln(0);
		if (!pos && player1.score > hs->score) {
			hiscoreline();
			pos = 1;
			continue;
		}
		putch(i+'0');
		putch('.');
		if (hs->score < 1000000)
			putch(' ');
		sprintf(s, "%06d", hs->score);
		printstr(s);
		hs++;
		i++;
	}
	if (!pos) {
		newln(0);
		if (ishiscore())
			hiscoreline();
		else
			cleartoeol();
	}
	refreshwin(WIN_TOP_SCORES);
}

static void setupgame(int i)
{
	allocgame(i & 2);
	initgame();
#ifndef NO_MENU
	if (!in_menu)
		goto skipmenu;
	while (term_gamemenu()) {
		in_menu = 0;
		rminvitfile();
skipmenu:
#endif
		writeconfig();
		term_entergame();
		while (startgame())
			;
#ifndef NO_MENU
		clearwin(0);
		in_menu = 1;
		term_entermenu();
		readoptions();
	}
	rmsocket();
	mkinvitfile();
#endif
	game = NULL;
}

void term_startup(int i)
{
	xterm_init();
	gettermoptions();
	term_setup();
	atexit(restoreterm);
	if (!bgdot)
		bgdot = default_bgdot();
#ifdef TWOPLAYER
	if (in_menu)
		term_entermenu();
	if (i)
		goto setup;
	i = getopt_int("", "mode");
	i = 2-!(i & MODE_2PLAYER);
	if (!in_menu)
		goto setup;
	mkinvitfile();
	while (1) {
# ifndef NO_MENU
		i = term_startupmenu(i);
		if (!i)
# endif
			break;
setup:		setupgame(i);
	}
#else /* if !TWOPLAYER */
	if (in_menu)
		term_entermenu();
	setupgame(1);
#endif
	restoreterm();
}
